"use strict";
//Code is basically same from https://github.com/texei/texei-sfdx-plugin
// Just updated it for the revised cli core
Object.defineProperty(exports, "__esModule", { value: true });
const command_1 = require("@salesforce/command");
const core_1 = require("@salesforce/core");
const spawn = require("child-process-promise").spawn;
const exec = require("child-process-promise").exec;
const packageIdPrefix = "0Ho";
const packageVersionIdPrefix = "04t";
const packageAliasesMap = [];
const defaultWait = 60;
// Initialize Messages with the current plugin directory
command_1.core.Messages.importMessagesDirectory(__dirname);
// Load the specific messages for this file. Messages from @salesforce/command, @salesforce/core,
// or any library that is using the messages framework can also be loaded this way.
const messages = command_1.core.Messages.loadMessages("sfpowerkit", "install");
let Install = /** @class */ (() => {
    class Install extends command_1.SfdxCommand {
        async run() {
            const result = { installedPackages: {} };
            // this.org is guaranteed because requiresUsername=true, as opposed to supportsUsername
            const username = this.org.getUsername();
            // Getting Project config
            const projectMain = await core_1.SfdxProject.resolve();
            const project = await projectMain.retrieveSfdxProjectJson();
            // Getting a list of alias
            const packageAliases = project.get("packageAliases") || {};
            if (typeof packageAliases !== undefined) {
                Object.entries(packageAliases).forEach(([key, value]) => {
                    packageAliasesMap[key] = value;
                });
            }
            //Validate Packages  installed in the target org
            let installedpackages = await this.getInstalledPackages(username);
            let individualpackage = this.flags.individualpackage;
            // Getting Package
            const packagesToInstall = [];
            const packageDirectories = project.get("packageDirectories") || [];
            // get branch filter
            this.branchMap = new Map();
            if (this.flags.branch) {
                this.branchMap = this.parseKeyValueMapfromString(this.flags.branch, "Branch", "core:branchname consumer:branchname packageN:branchname");
            }
            //get tag filter
            this.tagMap = new Map();
            if (this.flags.tag) {
                this.tagMap = this.parseKeyValueMapfromString(this.flags.tag, "Tag", "core:tag consumer:tag packageN:tag");
            }
            // get all packages in the mono repo project
            let monoRepoPackages = [];
            for (let packageDirectory of packageDirectories) {
                packageDirectory = packageDirectory;
                if (packageDirectory.path &&
                    packageDirectory.package &&
                    !monoRepoPackages.includes(packageDirectory.package.toString())) {
                    monoRepoPackages.push(packageDirectory.package.toString());
                }
            }
            for (let packageDirectory of packageDirectories) {
                packageDirectory = packageDirectory;
                const dependencies = packageDirectory.dependencies || [];
                if (this.flags.filterpaths &&
                    this.flags.filterpaths.length > 0 &&
                    !this.flags.filterpaths.includes(packageDirectory.path.toString())) {
                    continue;
                }
                if (dependencies && dependencies[0] !== undefined) {
                    this.ux.log(`\nPackage dependencies found for package directory ${packageDirectory.path}`);
                    for (const dependency of dependencies) {
                        const packageInfo = {};
                        const { package: dependentPackage, versionNumber } = dependency;
                        packageInfo.dependentPackage = dependentPackage;
                        packageInfo.versionNumber = versionNumber;
                        const packageVersionId = await this.getPackageVersionId(dependentPackage, versionNumber);
                        packageInfo.packageVersionId = packageVersionId;
                        if (individualpackage) {
                            if (packageInfo.dependentPackage.toString() === individualpackage ||
                                packageInfo.packageVersionId.toString() === individualpackage) {
                                packagesToInstall.push(packageInfo);
                                continue;
                            }
                        }
                        else {
                            if (this.flags.updateall) {
                                packagesToInstall.push(packageInfo);
                            }
                            else {
                                if (!installedpackages.includes(packageVersionId) &&
                                    !monoRepoPackages.includes(packageInfo.dependentPackage)) {
                                    packagesToInstall.push(packageInfo);
                                }
                            }
                        }
                        this.ux.log(`    ${packageInfo.packageVersionId} : ${packageInfo.dependentPackage}${packageInfo.versionNumber === undefined
                            ? ""
                            : " " + packageInfo.versionNumber}`);
                    }
                }
                else {
                    this.ux.log(`\nNo dependencies found for package directory ${packageDirectory.path}`);
                }
            }
            if (packagesToInstall.length > 0) {
                // Installing Packages
                let installationKeyMap = new Map();
                // Getting Installation Key(s)
                if (this.flags.installationkeys) {
                    installationKeyMap = this.parseKeyValueMapfromString(this.flags.installationkeys, "Installation Key", "core:key nCino:key vlocity:key");
                }
                this.ux.log("\n");
                for (let packageInfo of packagesToInstall) {
                    packageInfo = packageInfo;
                    if (result.installedPackages.hasOwnProperty(packageInfo.packageVersionId)) {
                        this.ux.log(`PackageVersionId ${packageInfo.packageVersionId} already installed. Skipping...`);
                        continue;
                    }
                    // Split arguments to use spawn
                    const args = [];
                    args.push("force:package:install");
                    // USERNAME
                    args.push("--targetusername");
                    args.push(`${username}`);
                    // PACKAGE ID
                    args.push("--package");
                    args.push(`${packageInfo.packageVersionId}`);
                    // INSTALLATION KEY
                    if (installationKeyMap &&
                        installationKeyMap.has(packageInfo.dependentPackage.toString())) {
                        let key = installationKeyMap.get(packageInfo.dependentPackage.toString());
                        args.push("-k");
                        args.push(`${key}`);
                    }
                    // WAIT
                    const wait = this.flags.wait ? this.flags.wait.trim() : defaultWait;
                    args.push("-w");
                    args.push(`${wait}`);
                    args.push("-b");
                    args.push(`${wait}`);
                    // NOPROMPT
                    if (this.flags.noprompt) {
                        args.push("--noprompt");
                    }
                    if (this.flags.apexcompileonlypackage) {
                        args.push("-a");
                        args.push(`package`);
                    }
                    this.ux.log(`Installing package ${packageInfo.packageVersionId} : ${packageInfo.dependentPackage}${packageInfo.versionNumber === undefined
                        ? ""
                        : " " + packageInfo.versionNumber}`);
                    var startTime = new Date().valueOf();
                    await spawn("sfdx", args, { stdio: "inherit" });
                    var endTime = new Date().valueOf();
                    var timeElapsed = (endTime - startTime) / 1000;
                    this.ux.log(`Elapsed time in installing package  ${packageInfo.packageVersionId} is ${timeElapsed} seconds`);
                    this.ux.log("\n");
                    result.installedPackages[packageInfo.packageVersionId] = packageInfo;
                }
            }
            else {
                this.ux.log("\n \n Looks like there is nothing to be updated in this org");
            }
            return { message: result };
        }
        async getPackageVersionId(name, version) {
            let packageId;
            // Keeping original name so that it can be used in error message if needed
            let packageName = name;
            // TODO: Some stuff are duplicated here, some code don't need to be executed for every package
            // First look if it's an alias
            if (typeof packageAliasesMap[packageName] !== "undefined") {
                packageName = packageAliasesMap[packageName];
            }
            if (packageName.startsWith(packageVersionIdPrefix)) {
                // Package2VersionId is set directly
                packageId = packageName;
            }
            else if (packageName.startsWith(packageIdPrefix)) {
                if (!version) {
                    throw new command_1.core.SfdxError(`version number is mandatory for ${name}`);
                }
                // Get Package version id from package + versionNumber
                const vers = version.split(".");
                let query = "Select SubscriberPackageVersionId, IsPasswordProtected, IsReleased ";
                query += "from Package2Version ";
                query += `where Package2Id='${packageName}' and MajorVersion=${vers[0]} and MinorVersion=${vers[1]} and PatchVersion=${vers[2]} `;
                // If Build Number isn't set to LATEST, look for the exact Package Version
                if (vers[3] !== "LATEST") {
                    query += `and BuildNumber=${vers[3]} `;
                }
                else if (this.flags.usedependencyvalidatedpackages) {
                    query += `and ValidationSkipped = false `;
                }
                // If Branch is specified, use it to filter
                if (this.flags.branch && this.branchMap.has(name)) {
                    query += `and Branch='${this.branchMap.get(name).trim()}' `;
                }
                // If tag is specified, use it to filter
                if (this.flags.tag && this.tagMap.has(name)) {
                    query += `and Tag='${this.tagMap.get(name).trim()}' `;
                }
                query += "ORDER BY BuildNumber,Createddate DESC Limit 1";
                // Query DevHub to get the expected Package2Version
                const conn = this.hubOrg.getConnection();
                const resultPackageId = (await conn.tooling.query(query));
                if (resultPackageId.size === 0) {
                    // Query returned no result
                    const errorMessage = `Unable to find SubscriberPackageVersionId for dependent package ${name}`;
                    throw new command_1.core.SfdxError(errorMessage);
                }
                else {
                    packageId = resultPackageId.records[0].SubscriberPackageVersionId;
                }
            }
            return packageId;
        }
        async getInstalledPackages(targetOrg) {
            let packages = [];
            let installedPackagesQuery = "SELECT Id, SubscriberPackageId, SubscriberPackage.NamespacePrefix, SubscriberPackage.Name, " +
                "SubscriberPackageVersion.Id, SubscriberPackageVersion.Name, SubscriberPackageVersion.MajorVersion, SubscriberPackageVersion.MinorVersion, " +
                "SubscriberPackageVersion.PatchVersion, SubscriberPackageVersion.BuildNumber FROM InstalledSubscriberPackage " +
                "ORDER BY SubscriberPackageId";
            const conn = this.org.getConnection();
            await conn.tooling.query(installedPackagesQuery).then(queryResult => {
                const records = queryResult.records;
                if (records && records.length > 0) {
                    this.ux.log(`Installed Packages in the org ${targetOrg}`);
                    const output = [];
                    records.forEach(record => {
                        packages.push(record["SubscriberPackageVersion"]["Id"]);
                        output.push({
                            name: record["SubscriberPackage"]["Name"],
                            package_version_name: record["SubscriberPackageVersion"]["Name"],
                            package_version_id: record["SubscriberPackageVersion"]["Id"],
                            versionNumber: `${record["SubscriberPackageVersion"]["MajorVersion"]}.${record["SubscriberPackageVersion"]["MinorVersion"]}.${record["SubscriberPackageVersion"]["PatchVersion"]}.${record["SubscriberPackageVersion"]["BuildNumber"]}`
                        });
                    });
                    this.ux.table(output, [
                        "name",
                        "package_version_name",
                        "package_version_id",
                        "versionNumber"
                    ]);
                }
            });
            return packages;
        }
        parseKeyValueMapfromString(request, item, format) {
            let response = new Map();
            request = request.trim();
            let requestList = request.split(" ");
            for (let element of requestList) {
                let packageNameWithValue = element.split(":");
                if (packageNameWithValue.length === 2) {
                    response.set(packageNameWithValue[0], packageNameWithValue[1]);
                }
                else {
                    // Format is not correct, throw an error
                    throw new command_1.core.SfdxError(`Error in parsing ${item}, format should be: ${format}`);
                }
            }
            return response;
        }
    }
    Install.description = messages.getMessage("commandDescription");
    Install.examples = [
        '$ sfdx sfpowerkit:package:dependencies:install -u MyScratchOrg -v MyDevHub -k "MyPackage1:Key MyPackage3:Key" -b "DEV"'
    ];
    Install.flagsConfig = {
        individualpackage: command_1.flags.string({
            char: "p",
            required: false,
            description: "Installs a specific package especially for upgrade scenario"
        }),
        installationkeys: command_1.flags.string({
            char: "k",
            required: false,
            description: "installation key for key-protected packages (format is packagename:key --> core:key nCino:key vlocity:key to allow some packages without installation key)"
        }),
        branch: command_1.flags.string({
            char: "b",
            required: false,
            description: "the package version’s branch (format is packagename:branchname --> core:branchname consumer:branchname packageN:branchname)"
        }),
        tag: command_1.flags.string({
            char: "t",
            required: false,
            description: "the package version’s tag (format is packagename:tag --> core:tag consumer:tag packageN:tag)"
        }),
        wait: command_1.flags.string({
            char: "w",
            required: false,
            description: "number of minutes to wait for installation status (also used for publishwait). Default is 10"
        }),
        noprompt: command_1.flags.boolean({
            char: "r",
            required: false,
            description: "allow Remote Site Settings and Content Security Policy websites to send or receive data without confirmation"
        }),
        updateall: command_1.flags.boolean({
            char: "o",
            required: false,
            description: "update all packages even if they are installed in the target org"
        }),
        apexcompileonlypackage: command_1.flags.boolean({
            char: "a",
            required: false,
            description: "compile the apex only in the package, by default only the compilation of the apex in the entire org is triggered"
        }),
        usedependencyvalidatedpackages: command_1.flags.boolean({
            required: false,
            description: "use dependency validated packages that matches the version number schema provide"
        }),
        filterpaths: command_1.flags.array({
            char: "f",
            required: false,
            description: "in a mono repo project, filter packageDirectories using path and install dependent packages only for the specified path"
        })
    };
    // Comment this out if your command does not require an org username
    Install.requiresUsername = true;
    // Comment this out if your command does not require a hub org username
    Install.requiresDevhubUsername = true;
    // Set this to true if your command requires a project workspace; 'requiresProject' is false by default
    Install.requiresProject = true;
    return Install;
})();
exports.default = Install;
//# sourceMappingURL=install.js.map